/*
byte2bin - 10. November 2002 - Matthias Gaertner - mgaert@web.de

This program can convert AS31 output files in byte format (-Fbyte)
in flat image files that are convenient for (older) EPROM burners.

Usage examples:
 as31 -Fbyte paulmon21.asm
 as31 -Fbyte extra.asm
 byte2bin -s 64K -o 32K paulmon.bin paulmon21.byte extra.byte

This would compile both files and produce an EPROM image for a
512 Kbit device where A15 is connected to VCC permanently.

 byte2bin -s 4K paulmon4.bin paulmon21.byte

This would produce an image file for a 4K EPROM.

Bit shuffling may be used in case the EPROM is not connected to the
bus using the normal 1:1 wiring. Assume the microprocessor's data
lines D7,..,D0 are connected to the memory data lines in the order
D6,..,D0,D7.

 byte2bin -D "65432107" paulmon.bin paulmon21.byte

could be used to create a suitable EPROM image file.


The AS31 "byte" format consists of individual lines, each containing
a 16-bit hexadecimal address, a colon, a space and an 8-bit hexadecimal
data byte. This code should cope with UNIX (LF) and DOS (CR,LF) line endings.

...
0F61: 85
0F62: 9B
0F63: 0D
0F64: 0E
...

Compiling this program is really simple:
 GCC:  gcc -o byte2bin byte2bin.c
 UNIX: make byte2bin
 BCC:  bcc32 -w-aus -ebyte2bin.exe byte2bin.c (**)
 MSVC: cl -o byte2bin.exe byte2bin.c

**: Borlandf BCC5.5 is available for free!
    See http://www.borland.com/bcppbuilder/freecompiler
  
byte2bin is in the public domain. It is distributed in the hope that it will
be useful, but without any warranty; without even the implied warranty of
merchantability or fitness for a particular purpose. 

Tab 4
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

typedef struct __options
{
	unsigned long lImageSize;
	unsigned long lOffset;
	unsigned char *pData;
	char *pShuffleA;
	char *pShuffleD;
	FILE *fIn;
} OPTIONS;

void help( void )
{
	printf(
"byte2bin - v1.00 - 10.11.2002 - Matthias Gaertner - mgaert@web.de\n"
"Usage:\n"
"byte2bin [-s <size>] [-o <offset>] [-u] <output file> {<input file>}+\n"
"\n"
"Example:\n"
"byte2bin -s 64K -o 32K paulmon.bin paulmon21.byte extra.byte\n"
"\n"
"byte2bin allocates an internal buffer of the given size (default: 32K)\n"
"and transfers all data in the given input file(s) into this buffer.\n"
"The input files are processed in sequence. Redundant definitions for the\n"
"same address in later files will overwrite earlier specifications.\n"
"\n"
"Options:\n"
" -s (size) Specify the image file size as a decimal number, optionally\n"
"    followed by K or M (i.e. 2048, 2K, 4M). The memory area will be\n"
"    initialized with 0xFF bytes.\n"
" -o (offset) Add an offset to all addresses of the input file. Default: 0.\n"
" -u (update) The 'output file' is read first, then modified, then written\n"
"    back. Use this to mix different input files at different offsets into\n"
"    one image. The file must contain at least the number of bytes specified\n"
"    by the -s option. Bytes beyond that will be ignored, but retained.\n"
"\n"
"Advanced options:\n"
" -A (Address shuffling) Specify a shuffling of the address bits. The 16\n"
"    input bits are denoted F-A9-0. A valid permutation is any 16-character\n"
"    sequence of these. Additionally, symbols H and L may be used to force\n"
"    address bits to high or low. Use this to allow for non-standard address\n"
"    line wirings (where uP's A0 is connected to memory's A8, for example).\n"
"    Shuffling is done on the input file addresses before adding the offset.\n"
"    Default is FEDCBA9876543210, i.e no shuffling.\n"
" -D (Data shuffling) Same as -A, but for the 8 data bits.\n"
"    Default is 76543210, i.e no shuffling.\n"
"\n"
"byte2bin is in the public domain. byte2bin is distributed in the hope that it\n"
"will be useful, but without any warranty; without even the implied warranty of\n"
"merchantability or fitness for a particular purpose.\n"
		);
}

static unsigned long shuffle( unsigned long d, int nBits, char *p )
{
	unsigned long r = 0;
	unsigned long n = (1<<(nBits-1));

	if( p == NULL )
	{
		return d;
	}
	while( nBits-- > 0 )
	{
		int bit = 0;

		if( *p == 'H' )
		{
			bit = 1;
		}
		else if( *p == 'L' )
		{
			bit = 0;
		}
		else
		{
			bit = ((d & (1 << *p)) != 0);
		}

		if( bit != 0 )
		{
			r |= n;
		}

		p++;
		n >>= 1;
	}
	return r;
}

static unsigned long shuffleA( unsigned long a, OPTIONS *o )
{
	unsigned long r;
	if( o->pShuffleA == NULL )
	{
		return a;
	}
	r = shuffle( a, 16, o->pShuffleA );
	return r;
}

static unsigned char shuffleD( unsigned char d, OPTIONS *o )
{
	unsigned long r;
	if( o->pShuffleD == NULL )
	{
		return d;
	}
	r = shuffle( (unsigned long)d, 8, o->pShuffleD );
	return (unsigned char)r;
}

static char decodexdigit( char c )
{
	if( c >= '0' && c <= '9' )
	{
		return c - '0';
	}
	if( c >= 'A' && c <= 'F' )
	{
		return c - 'A' + 10;
	}
	if( c >= 'a' && c <= 'f' )
	{
		return c - 'a' + 10;
	}
	return (char) -1;
}

static int read_byte( FILE *fIn )
{
	char c;
	char d;
	if( 1 != fread( &c, sizeof(c), 1, fIn ) || (c=decodexdigit(c)) < 0 )
	{
		return 1;
	}
	if( 1 != fread( &d, sizeof(d), 1, fIn ) || (d=decodexdigit(d)) < 0 )
	{
		return 1;
	}
	return c * 16 + d;
}

static int read_middle( FILE *fIn )
{
	char c;
	if( 1 != fread( &c, sizeof(c), 1, fIn ) || c != ':' )
	{
		return 1;
	}
	if( 1 != fread( &c, sizeof(c), 1, fIn ) || c != ' ' )
	{
		return 1;
	}
	return 0;
}

static int read_lineend( FILE *fIn )
{
	char c;
	if( 1 != fread( &c, sizeof(c), 1, fIn ) )
	{
		return 1;
	}
	if( c == '\r' )
	{
		if( 1 != fread( &c, sizeof(c), 1, fIn ) || c != '\n' )
		{
			return 1;
		}
	}
	else if( c != '\n' )
	{
		return 1;
	}
	return 0;
}

static int mix_in( OPTIONS *o )
{
	unsigned long nLine = 1;
	unsigned long a = 0;
	int c_high, c_low, c_data;

	while( !feof( o->fIn ) )
	{
		if( (c_high = read_byte(o->fIn)) < 0 ||
			(c_low = read_byte(o->fIn)) < 0 ||
			read_middle(o->fIn) < 0 ||
			(c_data = read_byte(o->fIn)) < 0 ||
			read_lineend(o->fIn) < 0 )
		{
			printf("\nError reading input file (line %lu)",nLine);
			return 1;
		}

		a = shuffleA((unsigned long)c_high * 256L + (unsigned long)c_low, o) +
			o->lOffset;
		if( a > o->lImageSize )
		{
			printf("\nAddress out of range (line %lu, 0x%06X > 0x%06X)",
				nLine, a, o->lImageSize );
			return 1;
		}
		c_data = (unsigned char)shuffleD((unsigned char)c_data,o);
		o->pData[a] = c_data;
		if( (++nLine % 256) == 0 )
		{
			printf(".");
		}
	}
	return 0;
}

static unsigned long scan_ulong( char *p )
{
	unsigned long n = 0;
	while( *p >= '0' && *p <= '9' )
	{
		n = n*10L + (*p++ - '0');
	}
	if( *p == 'K' || *p == 'k' )
	{
		n = n * 1024L;
		p++;
	}
	else if( *p == 'M' || *p == 'm' )
	{
		n = n * 1024L * 1024L;
		p++;
	}
	if( *p != '\0' )
	{
		return (unsigned long)-1;
	}
	return n;
}

static char *scan_shuffle( char *p, int n )
{
	int m = 0;
	char *q = p;
	char c;
	while( m < n )
	{
		if( *p == 'H' || *p == 'h' )
		{
			*p = 'H';
		}
		else if( *p == 'L' || *p == 'l' )
		{
			*p = 'L';
		}
		else if( (c=decodexdigit(*p)) >= 0 && c < (char)n )
		{
			/* valid shuffle character */
			*p = c;
		}
		else
		{
			return NULL;
		}
		m++;
		p++;
	}
	if( *p != '\0' )
	{
		return NULL;
	}
	return q;
}

int main( int argc, char **argv )
{
	int r = 1;
	FILE *fOut = NULL;
	char *strFileOut = NULL;
	char *strFileIn = NULL;

	int iNeedHelp = 1;
	int iUpdateMode = 0;
	int i = 0;
	unsigned long l = 0;
	OPTIONS o;

	o.lImageSize = 32768;	/* Default: 32K */
	o.lOffset = 0;          /* Default: 0 */
	o.pData = NULL;
	o.pShuffleA = NULL;
	o.pShuffleD = NULL;
	o.fIn = NULL;

	/* 1st step: read command line parameters, prepare operation */
	for(;;)
	{
		if( argc < 3 )
		{
			break;
		}

		/* Check command parameter: -s nnnn[K|M] */
		if( strcmp( argv[1], "-s" ) == 0 )
		{
			if( argc < 2 )
			{
				break;
			}
			argc -= 2;
			argv += 2;

			o.lImageSize = scan_ulong( argv[0] );
			if( ((long)o.lImageSize) <= 0 )
			{
				break;
			}
		}

		/* Check command parameter: -s nnnn[K|M] */
		else if( strcmp( argv[1], "-o" ) == 0 )
		{
			if( argc < 2 )
			{
				break;
			}
			argc -= 2;
			argv += 2;

			o.lOffset = scan_ulong( argv[0] );
			if( ((long)o.lOffset) < 0 )
			{
				break;
			}
		}

		/* Check command parameter: -A nnnn */
		else if( strcmp( argv[1], "-A" ) == 0 )
		{
			if( argc < 2 )
			{
				break;
			}
			argc -= 2;
			argv += 2;

			if( (o.pShuffleA = scan_shuffle( argv[0], 16 )) == NULL )
			{
				break;
			}
		}

		/* Check command parameter: -D nnnn */
		else if( strcmp( argv[1], "-D" ) == 0 )
		{
			if( argc < 2 )
			{
				break;
			}
			argc -= 2;
			argv += 2;

			if( (o.pShuffleD = scan_shuffle( argv[0], 8 )) == NULL )
			{
				break;
			}
		}

		/* Check command parameter: -u */
		else if( strcmp( argv[1], "-u" ) == 0 )
		{
			if( argc < 1 )
			{
				break;
			}
			argc--;
			argv++;

			iUpdateMode = 1;
		}

		if( argc < 3 )
		{
			break;
		}

		if( argv[1][0] == '-' )
		{
			continue;
		}

		iNeedHelp = 0;

		/* The fun starts here */
		o.pData = malloc( o.lImageSize );
		if( o.pData == NULL )
		{
			printf("Unable to allocate %lu bytes of memory for image buffer\n",
				o.lImageSize );
			break;
		}
		for( l = o.lImageSize; l > 0; )
		{
			 o.pData[--l] = (unsigned char)'\xFF';
		}

		/* Open output file for writing */
		strFileOut = argv[1];
		if( iUpdateMode != 0 )
		{
			/* Update mode: read 'output file' into buffer */
			fOut = fopen( strFileOut, "r+b" );
			if( fOut == NULL )
			{
				printf("Unable to open output file \"%s\" for update",
					strFileOut );
				break;
			}
			if( o.lImageSize != fread( o.pData, 1, o.lImageSize, fOut ) )
			{
				printf("Failed to read %lu from file \"%s\"",
					o.lImageSize, strFileOut );
				break;
			}
			fseek( fOut, 0L, SEEK_SET );
		}
		else
		{
			/* Normal mode: create output file */
			fOut = fopen( strFileOut, "wb" );
			if( fOut == NULL )
			{
				printf("Unable to create output file \"%s\"", strFileOut );
				break;
			}
		}
		argc--;
		argv++;

		r = 0;
		break;
	}

	/* 2nd step: read input files in sequence */
	while( r == 0 && fOut != NULL && argc > 1 )
	{
		/* Open current input file for reading */
		strFileIn = argv[1];
		o.fIn = fopen( strFileIn, "rb" );
		if( o.fIn == NULL )
		{
			printf("Unable to open output file \"%s\" for reading",
				strFileIn );
			r = 1;
			break;
		}
		argc--;
		argv++;

		printf("Mixing in \"%s\"", strFileIn );
		if( mix_in( &o ) != 0 )
		{
			r = 1;
			break;
		}
		printf("\n" );

		/* Close input file */
		fclose( o.fIn );
		o.fIn = NULL;

		if( argc <= 1 )
		{
			/* no more files present on command line */

			/* 3rd step: write data to output file */
			if( o.lImageSize != fwrite( o.pData, 1, o.lImageSize, fOut) )
			{
				printf("Error writing output file \"%s\"\n", strFileOut);
				r = 1;
			}
			else
			{
				printf("Done\n");
				strFileOut = NULL;
			}
			break;
		}
	}


	if( iNeedHelp != 0 )
	{
		help();
	}

	if( o.pData != NULL )
	{
		free( o.pData );
	}
	if( o.fIn != NULL )
	{
		fclose( o.fIn );
	}
	if( fOut != NULL )
	{
		fclose( fOut );
	}
	if( strFileOut != NULL && !iUpdateMode )
	{
		unlink( strFileOut );
	}
	return r;
}
